# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division

# compression detaults
cname = b'blosclz'
clevel = 5
shuffle = 1  # byte shuffle

# for persistence
metapath = '__zmeta__'
datapath = '__zdata__'
datasuffix = '.blosc'
