# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division


from zarr.core import empty, zeros, ones, full, array, open
from zarr.ext import blosc_version
from zarr import defaults
from zarr import constants
from zarr.version import version as __version__
